@@ -7,27 +7,27 @@ import json
 import jwt
 import os
 import requests
-from openapi_core import create_spec
-from openapi_core.schema.parameters.enums import ParameterLocation
 import yaml
-from . import errors
 import pickle
 import shutil
+import time
 import concurrent.futures
 import copy
-from typing import Dict, NewType, Mapping, Optional
+from typing import Dict, NewType, Mapping, Optional, Union
 from atomicwrites import atomic_write
-import openapi_core
+from tapipy.configuration import Config
 
+from tapipy.util import dereference_spec, retriable
 import tapipy.errors
+from . import errors
+
+start_time = time.time()
 
 # Type declarations
 ResourceName = NewType('ResourceName', str)
 ResourceUrl = NewType('ResourceUrl', str)
 SpecPath = NewType('SpecPath', str)
-OpenApiSpec = NewType('OpenApiSpec', openapi_core.schema.specs.models.Spec)
 Resources = Dict[ResourceName, ResourceUrl]
-Specs = Dict[ResourceName, OpenApiSpec]
 ResourceInfo = Mapping[ResourceName, SpecPath]
 
 
@@ -57,7 +57,8 @@ RESOURCES = {
         'jobs': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-jobs.yml",
         'apps': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-apps.yml",
         'workflows': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-workflows.yml",
-        'notifications': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-notifications.yml"
+        'notifications': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-notifications.yml",
+        'globus-proxy': f"local: {os.path.join(os.path.dirname(__file__), 'resources')}/openapi_v3-globus-proxy.yml"
     },
     'tapipy':{
         'actors': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-actors.yml',
@@ -70,11 +71,12 @@ RESOURCES = {
         'tenants': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-tenants.yml',
         'tokens': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-tokens.yml',
         'pgrest': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-pgrest.yml',
-        'pods': 'https://raw.githubusercontent.com/tapis-project/pods_service/prod/docs/openapi_v3-pods.yml',
+        'pods': 'https://raw.githubusercontent.com/tapis-project/tapipy/main/tapipy/resources/openapi_v3-pods.yml',
         'jobs': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-jobs.yml',
         'apps': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-apps.yml',
         'workflows': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-workflows.yml',
-        'notifications': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-notifications.yml'
+        'notifications': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-notifications.yml',
+        'globus-proxy': 'https://raw.githubusercontent.com/tapis-project/tapipy/prod/tapipy/resources/openapi_v3-globus-proxy.yml'
     },
     'prod': {
         'actors': 'https://raw.githubusercontent.com/TACC/abaco/prod-v3/docs/specs/openapi_v3.yml',               
@@ -91,7 +93,8 @@ RESOURCES = {
         'jobs': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/prod/jobs-client/src/main/resources/JobsAPI.yaml',
         'apps': 'https://raw.githubusercontent.com/tapis-project/openapi-apps/prod/AppsAPI.yaml',
         'workflows': 'https://raw.githubusercontent.com/tapis-project/tapis-workflows/prod/src/api/specs/WorkflowsAPI.yaml',
-        'notifications': 'https://raw.githubusercontent.com/tapis-project/openapi-notifications/prod/NotificationsAPI.yaml'
+        'notifications': 'https://raw.githubusercontent.com/tapis-project/openapi-notifications/prod/NotificationsAPI.yaml',
+        'globus-proxy': 'https://raw.githubusercontent.com/tapis-project/globus-proxy/prod/service/resources/openapi_v3.yml'
     },
     'staging': {
         'actors': 'https://raw.githubusercontent.com/TACC/abaco/dev-v3/docs/specs/openapi_v3.yml',               
@@ -108,7 +111,8 @@ RESOURCES = {
         'jobs': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/staging/jobs-client/src/main/resources/JobsAPI.yaml',
         'apps': 'https://raw.githubusercontent.com/tapis-project/openapi-apps/staging/AppsAPI.yaml',
         'workflows': 'https://raw.githubusercontent.com/tapis-project/tapis-workflows/stage/src/api/specs/WorkflowsAPI.yaml',
-        'notifications': 'https://raw.githubusercontent.com/tapis-project/openapi-notifications/staging/NotificationsAPI.yaml'
+        'notifications': 'https://raw.githubusercontent.com/tapis-project/openapi-notifications/staging/NotificationsAPI.yaml',
+        'globus-proxy': 'https://raw.githubusercontent.com/tapis-project/globus-proxy/staging/service/resources/openapi_v3.yml'
     },
     'dev': {
         'actors': 'https://raw.githubusercontent.com/TACC/abaco/dev-v3/docs/specs/openapi_v3.yml',
@@ -125,12 +129,13 @@ RESOURCES = {
         'jobs': 'https://raw.githubusercontent.com/tapis-project/tapis-client-java/dev/jobs-client/src/main/resources/JobsAPI.yaml',
         'apps': 'https://raw.githubusercontent.com/tapis-project/openapi-apps/dev/AppsAPI.yaml',
         'workflows': 'https://raw.githubusercontent.com/tapis-project/tapis-workflows/dev/src/api/specs/WorkflowsAPI.yaml',
-        'notifications': 'https://raw.githubusercontent.com/tapis-project/openapi-notifications/dev/NotificationsAPI.yaml'
+        'notifications': 'https://raw.githubusercontent.com/tapis-project/openapi-notifications/dev/NotificationsAPI.yaml',
+        'globus-proxy': 'https://raw.githubusercontent.com/tapis-project/globus-proxy/dev/service/resources/openapi_v3.yml'
     }
 }
 
 
-def _get_specs(resources: Resources, spec_dir: str = None, download_latest_specs: bool = False) -> Specs:
+def _get_specs(resources: Resources, spec_dir: str = None, download_latest_specs: bool = False):
     """
     Gets specs requested in resources.
     Will download any spec that is not already downloaded.
@@ -166,16 +171,15 @@ def download_and_pickle_spec_dicts(resources: Resources, spec_dir: str, download
     # Set off some parallel threads cause it's quick.
     # Switched from multiprocessing due to some startup warnings when import Tapipy in scripts
     if urls_to_download:
+        # This has a serious 1+ sec import, so we only import when needed.
+        global Spec
+        from openapi_core import Spec
         POOL_SIZE = os.environ.get('POOL_SIZE', 16)
         with concurrent.futures.ThreadPoolExecutor(max_workers=POOL_SIZE) as executor:
             executor.map(_thread_download_spec_dict, urls_to_download)
 
-    # Set off some parallel processes cause it's quick.
-    #POOL_SIZE = os.environ.get('POOL_SIZE', 16)
-    #pool = multiprocessing.Pool(processes=POOL_SIZE)
-    #pool.map(_thread_download_spec_dict, urls_to_download)
-    #pool.close()
-    #pool.join()
+
+
 
 
 def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
@@ -189,22 +193,19 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
     response = requests.get(resource_url)
     if response.status_code == 200:
         try:
-            # Loads yaml into Python dictionary
-            spec_dict = yaml.load(response.content, Loader=yaml.FullLoader)
-        except Exception as e:
-            print(f'Got exception when attempting to load yaml for '
-                  f'"{resource_name}" resource; exception: {e}')
-            return
-        try:
-            # Attempts to create spec from dict to ensure the spec is valid
-            # We do a deepcopy as create_spec for some reason adds fields
-            # to the dictionary that's given
-            test_spec_dict = copy.deepcopy(spec_dict)
-            create_spec(test_spec_dict)
+            # Directly creates spec from response. This validates spec as valid.
+            spec = Spec.from_file(response.content)
         except Exception as e:
             print(f'Got exception when test creating spec for "{resource_name}" '
                   f'resource; Spec probably not verifying; exception: {e}')
             return
+        try:
+            # Dereference spec so we have a dict we can pickle.
+            spec_dict = dereference_spec(spec)
+        except Exception as e:
+            print(f'Got exception when derefrencing spec for "{resource_name}" '
+                  f'resource; exception: {e}')
+            return
         try:
             # Pickles and saves the spec dict to the spec_path atomically
             with atomic_write(f'{spec_path}', overwrite=True, mode='wb') as spec_file:
@@ -218,7 +219,7 @@ def _thread_download_spec_dict(resource_info: ResourceInfo) -> None:
                        f'Did not get 200, got the following back:\n{response.text}')
 
 
-def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
+def unpickle_and_create_specs(resources: Resources, spec_dir: str):
     """
     Pickles loads a specifed spec_path and creates said spec.
     Can't be threaded, map doesn't allow spec object to be sent back.
@@ -228,25 +229,27 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
     # Get resource path to point the unpickling at.
     for resource_name, url in resources.items():
         if "local:" in url:
+            spec_path = url.replace('local:', '').strip()
             try:
-                spec_path = url.replace('local:', '').strip()
-                with open(spec_path, 'rb') as spec_file:
-                    spec_dict = yaml.load(spec_file, Loader=yaml.FullLoader)
-                specs.update({resource_name: create_spec(spec_dict)})
+                # This has a serious 1+ import, so we only import if needed.
+                from openapi_core import Spec
+                spec = Spec.from_file_path(spec_path)
+                spec_dict = dereference_spec(spec)
+                specs.update({resource_name: spec_dict})
                 dicts.update({resource_name: spec_dict})
             except Exception as e:
                 print(f'Error reading local "{resource_name}" resource. '
-                      f'Ensure path is absolute. e:{e}')
+                      f'Ensure path is absolute {spec_path}. e:{e}')
             continue
         _, _, spec_path = get_file_info_from_url(url, spec_dir)
         try:
             # Unpickle and create_spec
             with open(spec_path, 'rb') as spec_file:
                 spec_dict = pickle.load(spec_file)
-            specs.update({resource_name: create_spec(spec_dict)})
+            specs.update({resource_name: spec_dict})
             dicts.update({resource_name: spec_dict})
         except Exception as e:
-            print(f'Got exception trying to unpickle and create spec for '
+            print(f'Got exception trying to unpickle spec for '
                   f'spec_path: "{resource_name}"; exception: {e}')
             # Using resource name to look if the resource exists in the
             # tapipy prod resources. If it is found, we fall back and use that!
@@ -258,9 +261,9 @@ def unpickle_and_create_specs(resources: Resources, spec_dir: str) -> Specs:
                 try:
                     # Unpickle and create_spec
                     with open(spec_path, 'rb') as spec_file:
-                        spec_dict = pickle.load(spec_file)
-                    specs.update({resource_name: create_spec(spec_dict)})
-                    dicts.update({resource_name: spec_dict})
+                        spec = pickle.load(spec_file)
+                    specs.update({resource_name: spec})
+                    dicts.update({resource_name: spec})
                 except Exception as e:
                     print('Error opening tapipy prod spec. This is bad.')
             else:
@@ -294,8 +297,8 @@ def get_file_info_from_url(url: str, spec_dir: str):
                    .replace('/', '-')\
                    .lower()
     # Get directory and full name for spec file
-    full_spec_name = f'{spec_name}.pickle'
-    spec_path = f'{spec_dir}/{spec_name}.pickle'
+    full_spec_name = f'v2-{spec_name}.pickle'
+    spec_path = f'{spec_dir}/v2-{spec_name}.pickle'
     return spec_name, full_spec_name, spec_path
 
 
@@ -317,11 +320,8 @@ def get_spec_dir(spec_dir: str):
         # Fallback on the spec folder from the Tapipy package directory
         spec_dir = os.path.join(os.path.dirname(__file__), 'specs')
     return spec_dir
-    
-
 RESOURCE_SPECS, RESOURCE_DICTS = _get_specs(RESOURCES['tapipy'])
 
-
 def get_basic_auth_header(username: str, password: str) -> str:
     """
     Convenience function with will return a properly formatted Authorization
@@ -330,7 +330,6 @@ def get_basic_auth_header(username: str, password: str) -> str:
     user_pass = bytes(f"{username}:{password}", 'utf-8')
     return 'Basic {}'.format(b64encode(user_pass).decode())
 
-
 class Tenants(object):
     """
     Basic tenancy manager for Tapipy clients.
@@ -342,8 +341,9 @@ class Tenants(object):
 
     def reload_tenants(self):
         try:
-            sites = self.tapi_client.tenants.list_sites()
-            tenants = self.tapi_client.tenants.list_tenants()
+            # _config required to ensure user client-level config isn't applied during Tapipy initialization
+            sites = self.tapi_client.tenants.list_sites(_config=Config())
+            tenants = self.tapi_client.tenants.list_tenants(_config=Config())
         except Exception as e:
             raise errors.BaseTapyException(f"Unable to retrieve sites and tenants from the Tenants API. e: {e}")
         for t in tenants:
@@ -364,7 +364,6 @@ class Tenants(object):
             except KeyError:
                 raise errors.BaseTapyException("Tenant not found.")
 
-
 class Tapis(object):
     """
     A client for the Tapis API.
@@ -388,6 +387,7 @@ class Tapis(object):
                  tenants: Tenants = None,
                  debug_prints: bool = True,
                  resource_dicts: dict = {},
+                 config: Union[Config, dict] = Config(), # Provide a default configuration object
                  plugins = [], # list[str] <-- this type doesn't compile in python < 3.9
                  **kwargs
                  ):
@@ -483,7 +483,7 @@ class Tapis(object):
         # create resources for each API defined above. In the future we could make this more dynamic in multiple ways.
         for resource_name, spec in resource_specs.items():
             # each API is a top-level attribute on the DynaTapy object, a Resource object constructed as follows:
-            setattr(self, resource_name, Resource(resource_name, spec.paths, self))
+            setattr(self, resource_name, Resource(resource_name, spec['paths'], self))
 
         # deal with plugins
         self.plugins = plugins
@@ -496,6 +496,14 @@ class Tapis(object):
         # method signature should be def fn(op: Opertaion, response: Response, **kwargs)
         self.plugin_on_call_post_request_callables = []
 
+        self.config = config
+        if type(config) == dict:
+            self.config = Config(**config)
+        # Set the configuration object
+        if type(config) not in [Config, dict]:
+            raise TypeError("Tapis Client Config must be an instance of 'Config' or a dictionary")
+
+
         # we lazy-load the tenant_cache to prevent making a call to the Tenants API when not needed.
         if tenants:
             self.tenant_cache = tenants
@@ -508,6 +516,8 @@ class Tapis(object):
                 if t.base_url == base_url:
                     self.tenant_id = t.tenant_id
 
+        
+
         for p in plugins:
             # call each plugin's on_tapis_client_instantiation() function with all args passed in.
             importlib.import_module(p).on_tapis_client_instantiation(self,
@@ -803,10 +813,10 @@ class Tapis(object):
         r = requests.Request('POST',
                              url,
                              files={"file": open(source_file_path, 'rb')},
-                             _tapis_headers=headers).prepare()
+                             headers=headers).prepare()
         # call the plugins' pre-request callables:
-        for f in self.tapis_client.plugin_on_call_pre_request_callables:
-            f(self, r, **kwargs)
+        # for f in self.tapis_client.plugin_on_call_pre_request_callables:
+        #     f(self, r, **kwargs)
 
         # make the request and return the response object -
         try:
@@ -924,29 +934,29 @@ class Resource(object):
         # tapis_client stores configuration data (api_server, token, etc..)
         self.tapis_client = tapis_client
 
-        # Here we create an attr on the object for each operation_id in the spec. The attr is itself an
+        # Here we create an attr on the object for each operationId in the spec. The attr is itself an
         # Operation object, defined below, with a special __call__ method.
-        # Examples operation_id's inclue "list_files", "upload_file", etc...
+        # Examples operationId's include "list_files", "upload_file", etc...
         for path_name, path_desc in self.resource_spec.items():
             # Each path_desc is an openapi_core.schema.paths.models.Path object
             # it has an operations object, which is a dictionary of operations associated with the path:
-            for op, op_desc in path_desc.operations.items():
+            for op, op_desc in path_desc.items():
                 # each op_desc is an openapi_core.schema.operations.models.Operation object.
-                # the op_desc has a number of associated attributes, including operation_id, parameters, path_name, etc.
+                # the op_desc has a number of associated attributes, including operationId, parameters, etc.
                 # we create an Operation object for each one of these:
-                if not op_desc.operation_id:
-                    print(f"invalid op_dec for {resource_name}; missing operation_id. op_dec: {op_desc}")
+                if not op_desc["operationId"]:
+                    print(f"invalid op_dec for {resource_name}; missing operationId. op_desc: {op_desc}")
                     continue
-                setattr(self, op_desc.operation_id, Operation(self.resource_name, op_desc, self.tapis_client))
+                setattr(self, op_desc["operationId"], Operation(self.resource_name, op_desc, path_name, op, self.tapis_client))
 
 
 class Operation(object):
     """
     Represents a single operation on an API resource defined by an OpenAPI spec file.
-    Operation objects are in one-to-one correspondence with operation_id's defined in the spec file.
+    Operation objects are in one-to-one correspondence with operationId's defined in the spec file.
     """
 
-    def __init__(self, resource_name, op_desc, tapis_client):
+    def __init__(self, resource_name, op_desc, path_name, http_method, tapis_client):
         """
         Instantiate an operation. The op_desc should an openapi_core Operation object associated with the operation.
         :param resource_name: (str) The resource associated with this operation.
@@ -956,15 +966,17 @@ class Operation(object):
         """
         self.resource_name = resource_name
         self.op_desc = op_desc
+        self.path_name = path_name
+        self.http_method = http_method
         self.tapis_client = tapis_client
 
         # derived attributes - for convenience
-        self.operation_id = op_desc.operation_id
-        self.http_method = op_desc.http_method
-        self.path_parameters = [p for _, p in op_desc.parameters.items() if p.location == ParameterLocation.PATH]
-        self.query_parameters = [p for _, p in op_desc.parameters.items() if p.location == ParameterLocation.QUERY]
-        self.request_body = op_desc.request_body
+        self.operation_id = op_desc["operationId"]
+        self.path_parameters = [p for p in op_desc.get('parameters', []) if p['in'] == 'path']
+        self.query_parameters = [p for p in op_desc.get('parameters', []) if p['in'] == 'query']
+        self.request_body = op_desc.get('requestBody', {})
 
+    @retriable
     def __call__(self, **kwargs):
         """
         Turns the operation object into a callable. Arguments must be passed as kwargs, where the name of each kwarg
@@ -984,25 +996,26 @@ class Operation(object):
         # construct the http path -
         # some API definitions, such as SK, chose to not include the "/v3/" at the beginning of their paths,
         # so we add it in:
-        if not self.op_desc.path_name.startswith('/v3/'):
-            self.url = f'{base_url}/v3{self.op_desc.path_name}'  # base url
+        if not self.path_name.startswith('/v3/'):
+            self.url = f'{base_url}/v3{self.path_name}'  # base url
         else:
-            self.url = f'{base_url}{self.op_desc.path_name}'  # base url
+            self.url = f'{base_url}{self.path_name}'  # base url
+
         url = self.url
         for param in self.path_parameters:
             # look for the name in the kwargs
-            if param.required:
-                if param.name not in kwargs:
-                    raise errors.InvalidInputError(msg=f"{param.name} is a required argument.")
-            p_val = kwargs.pop(param.name)
+            if param.get("required"):
+                if param["name"] not in kwargs:
+                    raise errors.InvalidInputError(msg=f"{param['name']} is a required argument.")
+            p_val = kwargs.pop(param["name"])
             if isinstance(p_val, int):
                 p_val = str(p_val)
             if isinstance(p_val, float):
                 p_val = str(p_val)
-            if param.required and not p_val:
-                raise errors.InvalidInputError(msg=f"{param.name} is a required argument and cannot be None.")
+            if param.get("required") and not p_val:
+                raise errors.InvalidInputError(msg=f"{param['name']} is a required argument and cannot be None.")
             # replace the parameter in the path template with the parameter value
-            s = '{' + f'{param.name}' + '}'
+            s = '{' + f'{param["name"]}' + '}'
             url = url.replace(s, p_val)
 
         # check for the _tapis_debug flag for generating debug data
@@ -1017,13 +1030,13 @@ class Operation(object):
         params = {}
         for param in self.query_parameters:
             # look for the name in the kwargs
-            if param.required:
-                if param.name not in kwargs:
-                    raise errors.InvalidInputError(msg=f"{param.name} is a required argument.")
+            if param.get("required"):
+                if param["name"] not in kwargs:
+                    raise errors.InvalidInputError(msg=f"{param['name']} is a required argument.")
             # only set the parameter if it was actually sent in the function -
-            if param.name in kwargs:
-                p_val = kwargs.pop(param.name, None)
-                params[param.name] = p_val
+            if param["name"] in kwargs:
+                p_val = kwargs.pop(param["name"], None)
+                params[param["name"]] = p_val
         
         # allow arbitrary query parameters to b passed in via the special "_tapis_query_parameters" kwarg --
         if '_tapis_query_parameters' in kwargs:
@@ -1087,63 +1100,145 @@ class Operation(object):
                 raise errors.InvalidInputError(
                     msg="The headers argument, if passed, must be a dictionary-like object.")
 
+        # if X-Tapis-Tracking_ID (regardless of case) is in the headers we need to set tracking_id for validation
+        tracking_id = None
+        for k, v in headers.items():
+            if k.lower() == 'x-tapis-tracking-id' or k.lower() == 'x_tapis_tracking_id':
+                tracking_id = headers.pop(k)
+                break
+        if '_x_tapis_tracking_id' in kwargs:
+            if tracking_id:
+                raise errors.InvalidInputError(msg="The _x_tapis_tracking_id argument and the X-Tapis-Tracking-ID header cannot both be set.")
+            else:
+                tracking_id = kwargs.pop('_x_tapis_tracking_id')
+
+        # tracking_id header needs to be passed through to __call__ headers for splunk audit trails
+        if tracking_id:
+            try:
+                if not isinstance(tracking_id, str):
+                    raise errors.InvalidInputError(
+                        msg="The _x_tapis_tracking_id argument, if passed, must be a string.")
+
+                if not tracking_id.isascii():
+                    raise errors.InvalidInputError(
+                        msg="_x_tapis_tracking_id validation error. <namespace>.<unique identifier>. namespace is a non-empty, ASCII string of alphanumeric characters and underscores, followed be a single period, followed by an ASCII universally unique identifier string. Must be an entirely ASCII string.")
+
+                if len(tracking_id) > 126:
+                    raise errors.InvalidInputError(
+                        msg="_x_tapis_tracking_id validation error. <namespace>.<unique identifier>. namespace is a non-empty, ASCII string of alphanumeric characters and underscores, followed be a single period, followed by an ASCII universally unique identifier string. Must be less than 126 characters.")
+
+                # only one . in string
+                if tracking_id.count('.') != 1:
+                    raise errors.InvalidInputError(
+                        msg="_x_tapis_tracking_id validation error. <namespace>.<unique identifier>. namespace is a non-empty, ASCII string of alphanumeric characters and underscores, followed be a single period, followed by an ASCII universally unique identifier string. count('.') != 1.")
+
+                # ensure doesn't start or end with .
+                if tracking_id.startswith('.') or tracking_id.endswith('.'):
+                    raise errors.InvalidInputError(
+                        msg="_x_tapis_tracking_id validation error. <namespace>.<unique identifier>. namespace is a non-empty, ASCII string of alphanumeric characters and underscores, followed be a single period, followed by an ASCII universally unique identifier string. Cannot start or end with '.'.")
+
+                tracking_namespace, tracking_unique_identifier = tracking_id.split('.')
+                # check namespace is alphanumeric + underscores
+                if not all(c.isalnum() or c == '_' for c in tracking_namespace):
+                    raise errors.InvalidInputError(msg="Error: tracking_namespace contains invalid characters. Alphanumeric + underscores only.")
+
+                # check tracking_unique_identifier is alphanumeric + hyphens
+                if not all(c.isalnum() or c == '-' for c in tracking_unique_identifier):
+                    raise errors.InvalidInputError(msg="Error: tracking_unique_identifier contains invalid characters. Alphanumeric + hyphens only.")
+
+                headers.update({'X-Tapis-Tracking-ID': tracking_id})
+            except ValueError:
+                raise errors.InvalidInputError(
+                    msg=f"_x_tapis_tracking_id validation error. <namespace>.<unique identifier>. namespace is a non-empty, ASCII string of alphanumeric characters and underscores, followed be a single period, followed by an ASCII universally unique identifier string. Got x_tapis_tracking_id: {tracking_id}")
+
         # construct the data -
         data = None
+        files = None
         # these are the list of allowable request body content types; ex., 'application/json'.
-        if hasattr(self.op_desc.request_body, 'content') and hasattr(self.op_desc.request_body.content, 'keys'):
+        if 'content' in self.request_body and hasattr(self.request_body['content'], 'keys'):
             # Find possible headers, user specified
             # headers override spec's possibilities.
             if 'Content-Type' in headers:
-                requestContentTypes = headers['Content-Type']
+                request_content_types = headers['Content-Type']
             else:
-                requestContentTypes = self.op_desc.request_body.content.keys()
+                request_content_types = self.request_body['content'].keys()
 
             # Note: If multiple content types, we use first content type we have code for
-            if 'application/json' in requestContentTypes:
+            if 'application/json' in request_content_types:
                 headers['Content-Type'] = 'application/json'
-                required_fields = self.op_desc.request_body.content['application/json'].schema.required
+                required_fields = self.request_body['content']['application/json'].get('schema', {}).get('required', {})
+
                 data = {}
                 # if the request body has no defined properties, look for a single "request_body" parameter.
-                if self.op_desc.request_body.content['application/json'].schema.properties == {}:
+                if self.request_body['content']['application/json'].get('schema', {}).get('properties', {}) == {}:
                     # choice of "request_body" is arbitrary, as the property name is not provided by the
                     # openapi spec in this case
-                    data = kwargs['request_body']
+                    data = kwargs.get('request_body', {})
                 else:
                     # otherwise, the request body has defined properties, so look for each one in the function kwargs
-                    for p_name, p_desc in self.op_desc.request_body.content['application/json'].schema.properties.items():
+                    for p_name, p_desc in self.request_body['content']['application/json'].get('schema', {}).get('properties', {}).items():
                         if p_name in kwargs:
                             data[p_name] = kwargs[p_name]
                         elif p_name in required_fields:
                             raise errors.InvalidInputError(msg=f'{p_name} is a required argument.')
                     # serialize data before passing it to the request
                 data = json.dumps(data)
-            elif 'application/octet-stream' in requestContentTypes:
+            elif 'application/octet-stream' in request_content_types:
                 headers['Content-Type'] = 'application/octet-stream'
-                data = kwargs['message']
+                data = kwargs.get('message', {})
             # x-www-form-urlencoded
-            elif 'application/x-www-form-urlencoded' in requestContentTypes:
+            elif 'application/x-www-form-urlencoded' in request_content_types:
                 headers['Content-Type'] = 'application/x-www-form-urlencoded'
-                data = f"message={kwargs['message']}"
-            elif 'multipart/form-data' in requestContentTypes:
-                # todo - iterate over parts in self.op_desc.request_body.content['multipart/form-data'].schema.properties
-                raise NotImplementedError
+                data = f"message={kwargs.get('message', {})}"
+            elif 'multipart/form-data' in request_content_types:
+                # We DO NOT set multipart/form-data headers. The requests library will do header creation for us.
+                # multipart/form-data should use request.Request(files={"formpart1": "formdata1", "formpart2": "formdata2}, ...)
+                # With the files arg, requests will set the Content-Type, boundary and Content-Length headers for us.
+                # It doesn't seem to work if we set Content-Type ourselves.
+                # headers['Content-Type'] = 'multipart/form-data'
+
+                required_fields = self.request_body['content']['multipart/form-data'].get('schema', {}).get('required', {})
+                data = {}
+                # if the request body has no defined properties, look for a single "request_body" parameter.
+                if self.request_body['content']['multipart/form-data'].get('schema', {}).get('properties', {}) == {}:
+                    # choice of "request_body" is arbitrary, as the property name is not provided by the
+                    # openapi spec in this case
+                    data['request_body'] = kwargs.get('request_body', {})
+                else:
+                    # otherwise, the request body has defined properties, so look for each one in the function kwargs
+                    for p_name, p_desc in self.request_body['content']['multipart/form-data'].get('schema', {}).get('properties', {}).items():
+                        if p_name in kwargs:
+                            data[p_name] = kwargs[p_name]
+                        elif p_name in required_fields:
+                            raise errors.InvalidInputError(msg=f'{p_name} is a required argument.')
+                files = data
             else:
                 # We could default to providing message field
                 # Error seems better
                 # data = kwargs['message']
-                raise NotImplementedError(f'Content type/s specified have not been implemented. Types: {requestContentTypes}')
+                raise NotImplementedError(f'Content type/s specified have not been implemented. Types: {request_content_types}')
+
 
         # create a prepared request -
         # cf., https://requests.kennethreitz.org/en/master/user/advanced/#request-and-response-objects
-        r = requests.Request(http_method,
-                             url,
-                             params=params,
-                             data=data,
-                             headers=headers).prepare()
+        # Either multipart/form-data (files) or every other type of header.
+        if files:
+            r = requests.Request(http_method,
+                                 url,
+                                 params=params,
+                                 files=files,
+                                 headers=headers).prepare()
+        else:
+            r = requests.Request(http_method,
+                                 url,
+                                 params=params,
+                                 data=data,
+                                 headers=headers).prepare()
+
         # call the plugins' pre-request callables:
         for f in self.tapis_client.plugin_on_call_pre_request_callables:
             f(self, r, **kwargs)
-        
+
         # the create_token operation requires HTTP basic auth, though some services, such as the authenticator, need to
         # use create_token to generate tokens on behalf of other users; in these cases, it is important to not set the
         # BasicAuth header, so we look for a special kwarg in this case
@@ -1282,7 +1377,6 @@ class TapisResult(object):
                 setattr(self, 'result', [x for x in arg])
             else:
                 setattr(self, 'result', arg)
-
         for k, v in kwargs.items():
             # primitive types
             if type(v) in TapisResult.PRIMITIVE_TYPES:
